Object.defineProperty(HTMLMediaElement.prototype, 'playing', {
	get: function() {
		return !!(this.currentTime > 0 && !this.paused && !this.ended && this.readyState > 2);
	}
});

window.AudioContext = window.AudioContext || window.webkitAudioContext;

class AudioManager {
	constructor() {
		this.lines = {};
		this.context = undefined;
		this.fade_timers = {};
		this.bgm_meta_info = {};
		this.init_done = false;
	}

	init() {
		if (this.init_done) return;
		this.init_done = true;
		this.context = this.getCleanAudioContext();
		this.master_gain_node = this.context.createGain();
		this.master_gain_node.connect(this.context.destination);
		this.setGainValue(this.master_gain_node, 1);
	}

	setGainValue(gain_node, value) {
		if (!gain_node) return;
		if (!this.context) return;
		gain_node.gain.setValueAtTime(value, this.context.currentTime);
	}

	getCleanAudioContext() {
		var context = new AudioContext();
		var source = context.createBufferSource();
		source.connect(context.destination);
		source.start(0);

		var checkSampleRateBug = function(context) {
			if (context.sampleRate !== 44100) {
				context.close();
				context = new AudioContext();
				return true;
			}
			return false;
		};

		if (checkSampleRateBug(context)) {
			context = new AudioContext();
		}
		source.disconnect();
		return context;
	}

	fadeGain_(gain_node, gain_name, end_volume, duration, callback) {
		if (this.fade_timers[gain_name]) {
			var value_now = gain_node.gain.value;
			gain_node.gain.cancelScheduledValues(this.context.currentTime);
			this.setGainValue(gain_node, value_now);
			clearInterval(this.fade_timers[gain_name]);
			delete this.fade_timers[gain_name];
		}
		gain_node.gain.linearRampToValueAtTime(end_volume, this.context.currentTime + duration/1000);
		this.fade_timers[gain_name] = setTimeout(() => {
			clearInterval(this.fade_timers[gain_name]);
			delete this.fade_timers[gain_name];

			if (end_volume <= 0) {
				this.lineStop(this.lines[gain_name]);
			}
			if (typeof(callback) === "function") {
				callback();
			}
		}, duration);
	}


	fadeGain(gain_node, gain_name, end_volume, duration, callback) {
		var start_volume = gain_node.gain.value;
		var line = this.lines[gain_name];
		if (line && !line.bgm_playing) {
			this.setGainValue(gain_node, 0);
			start_volume = 0;
		}
		this.lineStart(line);

		if (duration <= 0 || !duration) {
			this.setGainValue(gain_node, end_volume);
			if (end_volume <= 0) {
				this.lineStop(line);
			}
		} else {
			var d_time = 100;
			var d = end_volume - start_volume;

			if (this.fade_timers[gain_name]) {
				clearInterval(this.fade_timers[gain_name]);
				delete this.fade_timers[gain_name];
			}
			var next_volume = start_volume;
			this.fade_timers[gain_name] = setInterval(() => {
				next_volume += d*d_time/duration;
				if (Math.abs(next_volume - start_volume) >= Math.abs(d)) {
					clearInterval(this.fade_timers[gain_name]);
					delete this.fade_timers[gain_name];

					this.setGainValue(gain_node, end_volume);
					if (end_volume <= 0) {
						this.lineStop(line);
					}
					if (typeof(callback) === "function") {
						callback();
					}
				} else {
					this.setGainValue(gain_node, next_volume);
				}
			}, d_time);
		}
	}

	suspend() {
		if (!this.context) return;

		this.fadeGain(this.master_gain_node, "_master", 0, 1000, () => {
			this.context.suspend();
		});
	}

	resume() {
		if (!this.context) return;

		this.context.resume();
		this.fadeGain(this.master_gain_node, "_master", 1, 1000);
	}

	getBinaryFile(filename) {
		return new Promise((resolve, reject) => {
			var xhr = new XMLHttpRequest();
			xhr.open("GET", filename, true);
			xhr.responseType = "arraybuffer";
			xhr.onload = function() {
				resolve(xhr.response);
			};
			xhr.onerror = function() {
				reject(xhr.status);
			};
			xhr.onabort = function() {
				reject(xhr.status);
			};
			xhr.ontimeout = function() {
				reject(xhr.status);
			};
			xhr.send();
		});
	}

	resetSource(filename) {
		var line = this.lines[filename];
		if (!line) return;

		var source = this.context.createBufferSource();
		source.buffer = line.source.buffer;
		source.connect(line.gain_node);
		line.source = source;
	}

	removeSource(filename) {
		var line = this.lines[filename];
		if (!line) return;

		delete this.lines[filename];
	}

	loadAudioFile(filename, is_bgm) {
		return new Promise((resolve, reject) => {
			if (!filename) {
				reject();
				return;
			}

			if (this.lines[filename]) {
				// audio already exists. reuse it.
				resolve(this.lines[filename]);
				return;
			}

			// use Web Audio API.
			this.getBinaryFile(filename).then((content) => {
				var line = {
					id: filename,
					gain_node: this.context.createGain(),
					source: this.context.createBufferSource(),
					bgm_playing: false,
					is_long: false,
					is_bgm: is_bgm,
				};
				this.context.decodeAudioData(content, (buffer) => {
					line.source.buffer = buffer;
					line.source.connect(line.gain_node);
					line.gain_node.connect(this.master_gain_node);
					this.setGainValue(line.gain_node, 0);
					this.lines[filename] = line;
					resolve(this.lines[filename]);
				});
			}).catch((status) => {
				console.error("failed to load audio file: " + filename);
				reject();
			});
		});
	}

	loadAudioFileLong(filename) {
		return new Promise((resolve, reject) => {
			if (!filename) {
				reject();
				return;
			}

			if (this.lines[filename]) {
				// audio already exists. reuse it.
				resolve(this.lines[filename]);
				return;
			}

			// use audio element.
			var audio = new Audio(filename);
			var line = {
				id: filename,
				gain_node: this.context.createGain(),
				source: this.context.createMediaElementSource(audio),
				bgm_playing: false,
				media_elem: audio,
				canplaythrough: false,
				is_long: true,
				is_bgm: true,
			};
			audio.addEventListener("canplaythrough", () => {
				if (line.canplaythrough) return;
				line.canplaythrough = true;
				line.source.connect(line.gain_node);
				line.gain_node.connect(this.master_gain_node);
				this.setGainValue(line.gain_node, 0);
				this.lines[filename] = line;
				resolve(this.lines[filename]);
			}, false);
			audio.load();
		});
	}

	playSE(filename, volume) {
		return new Promise((resolve, reject) => {
			this.loadAudioFile(filename, false).then((line) => {
				line.source.onended = function() {
					this.disconnect();
					resolve();
				};
				this.setGainValue(line.gain_node, isNaN(volume) ? 1 : volume);
				line.source.start(0);

				// for next
				this.resetSource(filename);
			}).catch((err) => {
				console.error(err);
				reject();
			});
		});
	}

	lineStart(line) {
		if (!line) return;
		if (line.bgm_playing) return;
		line.bgm_playing = true;
		if (line.media_elem) {
			line.media_elem.volume = 1;
			line.media_elem.currentTime = 0;
			line.media_elem.play();
		} else {
			line.source.start(0);
		}
	}

	lineStop(line) {
		if (!line) return;
		if (!line.bgm_playing) return;
		line.bgm_playing = false;
		if (line.media_elem) {
			line.media_elem.pause();
			this.removeSource(line.id);
		} else {
			line.source.stop(0);
			this.removeSource(line.id);
		}
	}

	fadeOutBGMAll(duration) {
		for (var i in this.lines) {
			var line = this.lines[i];
			if (!line.is_bgm) continue;
			if (line.is_long) {
				this.fadeBGMLong(line.id, 0, duration, line.media_elem.loop);
			} else {
				this.fadeBGM(line.id, 0, duration, line.source.loop);
			}
		}
	}

	fadeBGM(filename, volume, duration, loop, loop_start, loop_end) {
		return new Promise((resolve, reject) => {
			this.loadAudioFile(filename, true).then((line) => {
				line.source.onended = function() {
					this.disconnect();
					resolve();
				};

				line.source.loop = loop ? true : false;
				if (line.source.loop) {
					if (isNaN(loop_start)) {
						if (!this.bgm_meta_info[filename] || isNaN(this.bgm_meta_info[filename].loop_start)) {
							line.source.loopStart = 0;
						} else {
							line.source.loopStart = this.bgm_meta_info[filename].loop_start/1000;
						}
					} else {
						line.source.loopStart = loop_start/1000
					}
					if (isNaN(loop_end)) {
						if (!this.bgm_meta_info[filename] || isNaN(this.bgm_meta_info[filename].loop_end)) {
							line.source.loopEnd = 0;
						} else {
							line.source.loopEnd = this.bgm_meta_info[filename].loop_end/1000;
						}
					} else {
						line.source.loopEnd = loop_end/1000
					}
				}

				var end_volume = Math.min(Math.max(isNaN(volume) ? 1 : volume, 0), 1);
				duration = Math.max(isNaN(duration) ? 0 : duration, 0);
				this.fadeGain(line.gain_node, filename, end_volume, duration);
			}).catch((err) => {
				console.error(err);
				reject();
			});
		});
	}

	fadeBGMLong(filename, volume, duration, loop) {
		return new Promise((resolve, reject) => {
			this.loadAudioFileLong(filename).then((line) => {
				$(line.media_elem).off("ended");
				$(line.media_elem).on("ended", function() {
					line.bgm_playing = false;
					resolve();
				});

				var gain = line.gain_node.gain;

				// loop_start and/or loop_end is not available.
				line.media_elem.loop = loop ? true : false;

				var end_volume = Math.min(Math.max(isNaN(volume) ? 1 : volume, 0), 1);
				duration = Math.max(isNaN(duration) ? 0 : duration, 0);
				this.fadeGain(line.gain_node, filename, end_volume, duration);
			}).catch((err) => {
				console.error(err);
				reject();
			});
		});
	}

	setBGMMetaInfo(info) {
		if (!info.id) return;
		this.bgm_meta_info[info.id] = info;
	}
}
